package dk.brics.string.grammar;

import dk.brics.automaton.Automaton;
import dk.brics.string.directedgraph.StronglyConnectedComponents;
import dk.brics.string.grammar.operations.Component;
import dk.brics.string.grammar.operations.Component.Recursion;
import dk.brics.string.grammar.operations.AssertionCycleApproximation;
import dk.brics.string.grammar.operations.GrammarAsDirectedGraph;
import dk.brics.string.grammar.operations.OperationCycleApproximation;
import dk.brics.string.grammar.operations.RegularApproximation;
import dk.brics.string.stringoperations.BinaryOperation;
import dk.brics.string.stringoperations.UnaryOperation;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Context-free grammar with regular operations.
 *
 * @author Anders M&oslash;ller &lt;<a href="mailto:amoeller@cs.au.dk">amoeller@cs.au.dk</a>&gt;
 */
public class Grammar {
	/**
	 * @author Jn Ruhalovsk
	 */
	private int shiftKeyValue;
	private Map<String,dk.brics.string.grammar.Nonterminal> nt_map;
	
    private List<Nonterminal> nonterminals;

    /**
     * Constructs new empty grammar.
     */
    public Grammar() {
        nonterminals = new ArrayList<Nonterminal>();
    }

    /**
     * Adds new nonterminal.
     */
    public Nonterminal addNonterminal() {
        Nonterminal n = new Nonterminal(nonterminals.size());
        nonterminals.add(n);
        return n;
    }

    /**
     * Returns the (unmodifiable) list of nonterminals in this grammar.
     */
    public List<Nonterminal> getNonterminals() {
        return Collections.unmodifiableList(nonterminals);
    }

    /**
     * Adds new unit production [<tt>a -> b</tt>].
     * Ignored if <tt>a==b</tt>.
     */
    public void addUnitProduction(Nonterminal a, Nonterminal b) {
        if (a != b) {
            a.getProductions().add(new UnitProduction(b));
        }
    }

    /**
     * Adds new pair production [<tt>a -> b c</tt>].
     */
    public void addPairProduction(Nonterminal a, Nonterminal b, Nonterminal c) {
        a.getProductions().add(new PairProduction(b, c));
    }

    /**
     * Adds new automaton production [<tt>a -> reg</tt>].
     * Ignored if <tt>reg</tt> is empty.
     */
    public void addAutomatonProduction(Nonterminal a, Automaton reg) {
        if (!reg.isEmpty()) {
            a.getProductions().add(new AutomatonProduction(reg));
        }
    }

    /**
     * Adds new epsilon production [<tt>a -> ""</tt>].
     */
    public void addEpsilonProduction(Nonterminal a) {
        a.getProductions().add(new EpsilonProduction());
    }

    /**
     * Adds new unary operation production [<tt>a -> op1(b)</tt>].
     */
    public void addUnaryProduction(Nonterminal a, UnaryOperation op, Nonterminal b) {
        a.getProductions().add(new UnaryProduction(op, b));
    }

    /**
     * Adds new binary operation production [<tt>a -> op2(b,c)</tt>].
     */
    public void addBinaryProduction(Nonterminal a, BinaryOperation op, Nonterminal b, Nonterminal c) {
        a.getProductions().add(new BinaryProduction(op, b, c));
    }

    /**
     * Returns number of nonterminals in this grammar.
     */
    public int getNumberOfNonterminals() {
        return nonterminals.size();
    }

    /**
     * Returns number of productions in this grammar.
     */
    public int getNumberOfProductions() {
        int x = 0;
        for (Nonterminal n : nonterminals) {
            x += n.getProductions().size();
        }
        return x;
    }

    /**
     * Returns string representation of this grammar.
     */
    @Override
    public String toString() {
        StringBuilder s = new StringBuilder();
        for (Nonterminal v : nonterminals) {
            for (Production p : v.getProductions()) {
                s.append(v).append(" -> ").append(p).append("\n");
            }
        }
        return s.toString();
    }

    /**
     * Visits all productions with the given visitor.
     */
    public void visitProductions(ProductionVisitor v) {
        for (Nonterminal a : nonterminals) {
            for (Production p : a.getProductions()) {
                p.visitBy(a, v);
            }
        }
    }
    
    /**
     * Breaks operation cycles using {@link AssertionCycleApproximation} and then
     * approximates the remaining operation cycles with {@link OperationCycleApproximation}.
     */
    public void approximateOperationCycles() {
        new AssertionCycleApproximation(this).approximate();
        new OperationCycleApproximation(this).approximate();
    }

    /**
     * See {@link OperationCycleApproximation#countCycles()}.
     */
    public int getNumberOfOperationCycles() {
        return new OperationCycleApproximation(this).countCycles();
    }

    /**
     * Finds the strongly connected components and their recursion kinds.
     */
    public StronglyConnectedComponents<Nonterminal, Component> getComponents(boolean find_recursion_kinds) {
        StronglyConnectedComponents<Nonterminal, Component> comp =
                new StronglyConnectedComponents<Nonterminal, Component>(new GrammarAsDirectedGraph(this));
        if (find_recursion_kinds) {
            for (Component c : comp.getComponents()) {
            	for (Nonterminal n : c.getNodes()) {
            		System.out.print(n.toString() + "->PRODS:" + n.getProductions().toString() + " | " );
            	}
            	System.out.println();
            	//System.out.println("C.getNode(0): " + c.getNode(0).getKey() + " PRODS: " + c.getNode(0).getProductions().toString());
            	//System.out.println("JANKO: " + comp.getComponent(this.getNonterminal(18)).getNode(0).toString() + " PRODU: " +  comp.getComponent(this.getNonterminal(18)).getNode(0).getProductions());
                c.findRecursion();
            }
        }
        return comp;
    }

    /**
     * Returns number of components.
     */
    public int getNumberOfComponents() {
        return getComponents(false).getComponents().size();
    }

    /**
     * Returns number of non-linear components.
     */
    public int getNumberOfNonLinearComponents() {
        int nonreg = 0;
        StronglyConnectedComponents<Nonterminal, Component> comp = getComponents(true);
        for (Component c : comp.getComponents()) {
            if (c.getRecursion() == Recursion.BOTH) {
                nonreg++;
            }
        }
        return nonreg;
    }

    /**
     * See {@link OperationCycleApproximation#getCharsets()}.
     */
    public String getCharsets() {
        return new OperationCycleApproximation(this).getCharsets();
    }

    /**
     * See {@link RegularApproximation}.
     */
    public void approximateNonLinear(Collection<Nonterminal> hotspots) {
        new RegularApproximation(this).approximate(hotspots);
    }
    
    /**
     * Constructs new grammar with nonterminal and their productions.
     * @author Jn Ruhalovsk
     */
    public Grammar(List<Nonterminal> n) {
    	nonterminals = new ArrayList<Nonterminal>();
        nonterminals.addAll(n);
    }
    
    /**
     * remove production for nonterminal, return productions for nonterminal(lhs) and productions(rhs)
     * if don't exist, return null
     * @author Jn Ruhalovsk
     */
    public Production removeProduction(String nameOfNonterminal, Production p) {
    	//System.out.println("NON: " + nameOfNonterminal + " PRO: " + p.toString());
    	for (int i=0; i<this.getNumberOfNonterminals(); i++) {
    		if (nameOfNonterminal.equals(this.getNonterminals().get(i).toString()) ) {
    			//for (int j=0; j<this.getNonterminals().get(i).getProductions().size(); j++) {
    			for (int j=0; j<this.getNonterminals().get(i).getProductions().size(); j++) {
    				//System.out.println("H> " + this.getNonterminals().get(i).getProductions().get(j).toString());
    				if (p.toString().equals(this.getNonterminals().get(i).getProductions().get(j).toString())) {
    					return this.getNonterminals().get(i).getProductions().remove(j);
    					
    				}
    			}
    			//return this.getNonterminals().get(i);
    		}
    	}
    	return null;
    }
    
    /**
     * getter for nonterminal, return productions for index of nonterminal
     * @author Jn Ruhalovsk
     */
    public Nonterminal getNonterminal(String nameOfNonterminal) {
    	for (int i=0; i<this.getNumberOfNonterminals(); i++) {
    		if (nameOfNonterminal.equals(this.getNonterminals().get(i).toString()) ) {
    			return this.getNonterminals().get(i);
    		}
    	}
    	return null;
    }
    
    /**
     * getter for nonterminal, return productions for nonterminal
     * @author Jn Ruhalovsk
     */
    public Nonterminal getNonterminal(int key) {
    	for (Nonterminal nn : this.getNonterminals()) {
    		if (nn.getKey() == key) {
    			return nn;
    		}
    	}
    	return null;
    }
    
    /**
     * getter for nonterminal, return productions for nonterminal
     * @author Jn Ruhalovsk
     */
    public Nonterminal getNonterminal(Nonterminal n) {
    	for (Nonterminal nn : this.getNonterminals()) {
    		if (nn.toString().equals(n.toString())) {
    			return nn;
    		}
    	}
    	return null;
    }
    
    /**
     * getter for productions, return productions for nonterminal(lhs) and productions(rhs)
     * if don't exist, return null
     * @author Jn Ruhalovsk
     */
    public Production getProduction(String nameOfNonterminal, Production p) {
    	for (int i=0; i<this.getNumberOfNonterminals(); i++) {
    		if (nameOfNonterminal.equals(this.getNonterminals().get(i).toString()) ) {
    			//for (int j=0; j<this.getNonterminals().get(i).getProductions().size(); j++) {
    			for (Production pp : this.getNonterminals().get(i).getProductions()) {
    				if (pp.toString().equals(p.toString())) {
    					return pp;
    				}
    			}
    			//return this.getNonterminals().get(i);
    		}
    	}
    	return null;
    }
    
    /**
     * shift key values
     * @author Jn Ruhalovsk
     */
    public void shiftKeyValuesForNonterminals(int constShiftKey) {
    	shiftKeyValue = constShiftKey;
    	for (Nonterminal n : nonterminals) {
    		n.setKey(shiftKeyValue + n.getKey());
    		for (Production p : n.getProductions()) {
    			//if (notShifted) {
    				//make shift on rhs
    			//}
    		}
    	}
    }
    
    /**
     * getter for shif key value
     * @author Jn Ruhalovsk
     * @return
     */
    public int getShiftKeyValue() {
    	return shiftKeyValue;
    }
    
    /**
     * addProduction from other grammar
     * @author Jn Ruhalovsk
     */
    public void addProductionsFromGrammar(Grammar g) {
    	 nt_map = new HashMap<String,dk.brics.string.grammar.Nonterminal>();
    	 for (Nonterminal n : g.getNonterminals()) {
    		this.addNonterminal(n.getKey());
    		System.out.println("VELKOST PRAVEJ STRANY pre " + n.toString() + " : " + n.getProductions().size());
        	for (Production p : n.getProductions()) {
        		//Nonterminal nn = new Nonterminal(n.getKey());
        		Nonterminal nn;
        		if ( this.getNonterminal(n.getKey()) != null) {
        			nn = this.getNonterminal(n.getKey()); 
        		}
        		else {
        			nn = this.addNonterminal(n.getKey());
        		}
        		//Nonterminal n1 =
        		//System.out.println("NN: " + nn);
        		if (p instanceof UnitProduction) {
        			Nonterminal n1 = ((UnitProduction) p).getNonterminal();
        			if ( this.getNonterminal(n1.getKey()) != null ) { //if was added so nonterminal
        				this.addUnitProduction(nn, this.getNonterminal(n1.getKey()) );
        			}
        			else {
        				this.addUnitProduction(nn, new Nonterminal(n1.getKey()) );
        			}
        			if (!nt_map.containsKey(n)) {
        				nt_map.put(n.toString(), n);
        			}
        		}
        		else if (p instanceof UnaryProduction) {
        			Nonterminal n1 = ((UnaryProduction) p).getNonterminal();
        			if ( this.getNonterminal(n1.getKey()) != null ) { //if was added so nonterminal
        				this.addUnaryProduction(nn, ((UnaryProduction) p).getOperation(), this.getNonterminal(n1.getKey()) );
        			}
        			else {
        				this.addUnaryProduction(nn, ((UnaryProduction) p).getOperation(), new Nonterminal(n1.getKey()) );
        			}
        			if (!nt_map.containsKey(n)) {
        				nt_map.put(n.toString(), n);
        			}
        		}
        		else if (p instanceof BinaryProduction) {
        			Nonterminal n1 = ((BinaryProduction) p).getNonterminal1();
        			Nonterminal n2 = ((BinaryProduction) p).getNonterminal2();
        			if ( this.getNonterminal(n1.getKey()) == null && this.getNonterminal(n2.getKey()) == null ) {
        				this.addBinaryProduction(nn, ((BinaryProduction) p).getOperation(), new Nonterminal(n1.getKey()), new Nonterminal(n2.getKey()));
        			}
        			else if ( this.getNonterminal(n1.getKey()) == null ) {
        				this.addBinaryProduction(nn, ((BinaryProduction) p).getOperation(), new Nonterminal(n1.getKey()), this.getNonterminal(n2.getKey()));
        			}
        			else if ( this.getNonterminal(n2.getKey()) == null) {
        				this.addBinaryProduction(nn, ((BinaryProduction) p).getOperation(), this.getNonterminal(n2.getKey()), new Nonterminal(n2.getKey()));
        			}
        			else {
        				this.addBinaryProduction(nn, ((BinaryProduction) p).getOperation(), this.getNonterminal(n2.getKey()), this.getNonterminal(n2.getKey()));
        			}
        			if (!nt_map.containsKey(n)) {
        				nt_map.put(n.toString(), n);
        			}
        		}
        		else if (p instanceof AutomatonProduction) {
        			this.addAutomatonProduction(nn, ((AutomatonProduction) p).getAutomaton());
        		}
        		else if (p instanceof EpsilonProduction) {
        			this.addEpsilonProduction(nn);
        		}
        		else if (p instanceof PairProduction) {
        			System.out.println("PAIR in GRAMMAR.class: " + p.toString());
        			Nonterminal n1 = ((PairProduction) p).getNonterminal1();
        			Nonterminal n2 = ((PairProduction) p).getNonterminal2();
        			if ( this.getNonterminal(n1.getKey()) == null && this.getNonterminal(n2.getKey()) == null ) {
        				this.addPairProduction(nn, new Nonterminal(n1.getKey()), new Nonterminal(n2.getKey()));
        				System.out.println("FIRST");
        			}
        			else if ( this.getNonterminal(n1.getKey()) == null ) {
        				this.addPairProduction(nn, new Nonterminal(n1.getKey()), this.getNonterminal(n2.getKey()));
        				System.out.println("SECOND");
        			}
        			else if ( this.getNonterminal(n2.getKey()) == null) {
        				this.addPairProduction(nn, this.getNonterminal(n1.getKey()), new Nonterminal(n2.getKey()));
        				System.out.println("THIRD");
        			}
        			else {
        				this.addPairProduction(nn, this.getNonterminal(n1.getKey()), this.getNonterminal(n2.getKey()));
        			}
        			if (!nt_map.containsKey(n)) {
        				nt_map.put(n.toString(), n);
        			}
        			//this.addPairProduction(nn, ((PairProduction) p).getNonterminal1(), ((PairProduction) p).getNonterminal2());
        		}
        	}
		}
    	 System.out.println("GRammar.class: Nt_MAP: " + nt_map.toString() );
    }
    
    /**
     * Adds new nonterminal.
     * @author Jn Ruhalovsk
     */
    public Nonterminal addNonterminal(int key) {
        Nonterminal n = new Nonterminal(key);
        nonterminals.add(n);
        return n;
    }
    
    /**
     * returnmapped nonterminal after shifting
     * @author Jn Ruhalovsk
     */
    public Map<String,dk.brics.string.grammar.Nonterminal> gettShiftedNonterminalMap () {
    	return nt_map;
    }
}
