package dk.brics.string.grammar.operations;

import dk.brics.string.directedgraph.StronglyConnectedComponents;
import dk.brics.string.grammar.*;
import dk.brics.string.grammar.operations.Component.Recursion;
import dk.brics.string.mlfa.MLFA;
import dk.brics.string.mlfa.MLFAState;
import dk.brics.string.mlfa.MLFAStatePair;

/**
 * Converter from strongly regular grammars to MLFAs.
 *
 * @see dk.brics.string.StringAnalysis
 */
public class Grammar2MLFA {

    private Grammar g;

    private StronglyConnectedComponents<Nonterminal, Component> comp;

    private MLFAState[] nt_states;

    /**
     * Constructs a new converter for the given grammar.
     */
    public Grammar2MLFA(Grammar g) {
        this.g = g;
    }

    /**
     * Constructs {@link MLFA} from the grammar.
     * The grammar is assumed to be approximated first (see {@link RegularApproximation}).
     *
     * @see #getMLFAStatePair(Nonterminal)
     */
    public MLFA convert() {
        comp = g.getComponents(true);
        //System.out.println("TEST NODE: " + comp.getComponent(g.getNonterminal(18)) + " PRODUCTions: " + comp.getComponent(g.getNonterminal(18)).getNode(0).getProductions());
        System.out.println("COMPONENTS: " + comp.toString() );
        nt_states = new MLFAState[g.getNumberOfNonterminals()];
        MLFA mlfa = new MLFA();
        
        for (Component c : comp.getComponents()) {
        	System.out.println("==================convertCOMPONENT========================");
        	//System.out.println("MLFA convert: " + c.getNode(0).getProductions());
            convertComponent(c, mlfa);
            //System.out.println("COMPONENT: " + c.toString());
        }
        return mlfa;
    }

    /**
     * Returns MLFA state pair representing initial and final state for the given nonterminal.
     * {@link #convert()} must be called first.
     */
    public MLFAStatePair getMLFAStatePair(Nonterminal n) {
        MLFAStatePair p;
        //System.out.println("grammat2mlfa.class: "+ comp.toString());
        Component c = comp.getComponent(n);
        if (c.getRecursion() == Recursion.RIGHT || c.getRecursion() == Recursion.NONE) {
        	//System.out.println("for nonterminal with key: " + n.getKey() + " initital state:" + nt_states[n.getKey()] + " final state: " + c.getState().toString());
            p = new MLFAStatePair(nt_states[n.getKey()], c.getState());
        } else {
        	//System.out.println("c.state2: " + c.getState().toString() + " nt_states[n.getKey]: " +nt_states[n.getKey()].toString());
        	//System.out.println("for nonterminal with key: " + n.getKey() + " initital state:" + c.getState().toString() + " final state: " + nt_states[n.getKey()]);
            p = new MLFAStatePair(c.getState(), nt_states[n.getKey()]);
        }
        p.setTaint(n.isTaint());
        return p;
    }

    private void convertComponent(final Component c, final MLFA mlfa) {
        if (c.getState() == null) {
        	
            // make states
            c.setState(mlfa.addState());
            //System.out.println("STATE KEY: " + c.getState().getKey());
            for (Nonterminal a : c.getNodes()) {
            	//System.out.println("Nt_STATES:" + a.getKey() + " SIZE OF PRODUCTIONS: " + a.getProductions().size());
                nt_states[a.getKey()] = mlfa.addState();
                //System.out.println("ADDED MLFASTATE: " + nt_states[a.getKey()].toString());
            }
            
            //System.out.println("NODES in GRAMMAR2MLFA: " + c.getNodes().toString());

            // make transitions
            for (Nonterminal a : c.getNodes()) {
            	System.out.println("NONTERMINAL: " + a.toString() + " SIZE OF PRODUCTIONS: " + a.getProductions().size());
//            	//added by Jn Ruhalovsk
//            	if (a.getProductions().size() == 0) {
//            		a.getProductions().addAll(g.getNonterminal(a).getProductions());
//            	}
            	//System.out.println("NONTERMINAL AFTER: " + a.toString() + " SIZE OF PRODUCTIONS: " + a.getProductions().size());
                for (Production p : a.getProductions()) {
                    if (c.getRecursion() == Recursion.BOTH) {
                        throw new RuntimeException("grammar is not strongly regular");
                    } else if (c.getRecursion() == Recursion.RIGHT || c.getRecursion() == Recursion.NONE) {
                        // component is right or non recursive
                        p.visitBy(a, new ProductionVisitor() {

                            public void visitUnitProduction(Nonterminal a, UnitProduction p) {
                            	//System.out.println("MLAF UnitPRODUCTION: " + a.toString() + "prod: " + p.toString());
                                if (c.contains(p.getNonterminal())) {
                                    // A -> B  =>  (A)--e-->(B)
                                    mlfa.addEpsilonTransition(nt_states[a.getKey()], nt_states[p.getNonterminal().getKey()]);
                                } else {
                                    // A -> X  =>  (A)--[X]-->(T)
                                    convertComponent(comp.getComponent(p.getNonterminal()), mlfa);
                                    mlfa.addIdentityTransition(nt_states[a.getKey()], c.getState(), getMLFAStatePair(p.getNonterminal()));
                                }
                            }

                            public void visitPairProduction(Nonterminal a, PairProduction p) {
                            	//System.out.println("MLAF PairPRODUCTION: " + a.toString() + "prod: " + p.toString());
                                if (c.contains(p.getNonterminal2())) {
                                    // A -> X B  =>  (A)--[X]-->(B)
                                    convertComponent(comp.getComponent(p.getNonterminal1()), mlfa);
                                    mlfa.addIdentityTransition(nt_states[a.getKey()], nt_states[p.getNonterminal2().getKey()], getMLFAStatePair(p.getNonterminal1()));
                                } else {
                                    // A -> X Y  =>  (A)--[X]-->(R), (R)--[Y]-->(T)
                                    MLFAState r = mlfa.addState();
                                    convertComponent(comp.getComponent(p.getNonterminal1()), mlfa);
                                    convertComponent(comp.getComponent(p.getNonterminal2()), mlfa);
                                    mlfa.addIdentityTransition(nt_states[a.getKey()], r, getMLFAStatePair(p.getNonterminal1()));
                                    mlfa.addIdentityTransition(r, c.getState(), getMLFAStatePair(p.getNonterminal2()));
                                }
                            }

                            public void visitAutomatonProduction(Nonterminal a, AutomatonProduction p) {
                            	//System.out.println("MLAF AtomatonPRODUCTION: " + a.toString() + "prod: " + p.toString());
                                // A -> reg  =>  (A)--[reg]-->(T)
                                mlfa.addAutomatonTransition(nt_states[a.getKey()], c.getState(), p.getAutomaton());
                            }

                            public void visitUnaryProduction(Nonterminal a, UnaryProduction p) {
                            	//System.out.println("MLAF UnaryPRODUCTION: " + a.toString() + "prod: " + p.toString());
                                // A -> op1(X)  =>  (A)--[op1(X)]-->(T)
                                convertComponent(comp.getComponent(p.getNonterminal()), mlfa);
                                mlfa.addUnaryTransition(nt_states[a.getKey()], c.getState(), p.getOperation(), getMLFAStatePair(p.getNonterminal()));
                            }

                            public void visitBinaryProduction(Nonterminal a, BinaryProduction p) {
                            	//System.out.println("MLAF BinaryPRODUCTION: " + a.toString() + "prod: " + p.toString());
                                // A -> op2(X,Y)  =>  (A)--[op2(X,Y)]-->(T)
                                convertComponent(comp.getComponent(p.getNonterminal1()), mlfa);
                                convertComponent(comp.getComponent(p.getNonterminal2()), mlfa);
                                mlfa.addBinaryTransition(nt_states[a.getKey()], c.getState(), p.getOperation(),
                                        getMLFAStatePair(p.getNonterminal1()), getMLFAStatePair(p.getNonterminal2()));
                            }

                            public void visitEpsilonProduction(Nonterminal a, EpsilonProduction p) {
                            	//System.out.println("MLAF EpsilonPRODUCTION: " + a.toString() + "prod: " + p.toString());
                                // A -> ""  =>  (A)--e-->(T)
                                mlfa.addEpsilonTransition(nt_states[a.getKey()], c.getState());
                            }
                        });
                    } else {
                        // component is left recursive
                        p.visitBy(a, new ProductionVisitor() {

                            public void visitUnitProduction(Nonterminal a, UnitProduction p) {
                            	//System.out.println("LEFT MLAF UnitPRODUCTION: " + a.toString() + " prod: " + p.toString());
                                if (c.contains(p.getNonterminal())) {                   
                                    // A -> B  =>  (B)---e--->(A)
                                    mlfa.addEpsilonTransition(nt_states[p.getNonterminal().getKey()], nt_states[a.getKey()]);
                                } else {
                                    // A -> X  =>  (S)--[X]-->(A)
                                    convertComponent(comp.getComponent(p.getNonterminal()), mlfa);
                                    mlfa.addIdentityTransition(c.getState(), nt_states[a.getKey()], getMLFAStatePair(p.getNonterminal()));
                                }
                            }

                            public void visitPairProduction(Nonterminal a, PairProduction p) {
                            	//System.out.println("LEFT MLAF PairPRODUCTION: " + a.toString() + " prod: " + p.toString());
                                if (c.contains(p.getNonterminal1())) {
                                    // A -> B X  =>  (B)--[X]-->(A)
                                    convertComponent(comp.getComponent(p.getNonterminal2()), mlfa);
                                    mlfa.addIdentityTransition(nt_states[p.getNonterminal1().getKey()], nt_states[a.getKey()], getMLFAStatePair(p.getNonterminal2()));
                                } else {
                                    // A -> X Y  =>  (S)--[X]-->(R), (R)--[Y]-->(A)
                                    MLFAState r = mlfa.addState();
                                    convertComponent(comp.getComponent(p.getNonterminal1()), mlfa);
                                    convertComponent(comp.getComponent(p.getNonterminal2()), mlfa);
                                    mlfa.addIdentityTransition(c.getState(), r, getMLFAStatePair(p.getNonterminal1()));
                                    mlfa.addIdentityTransition(r, nt_states[a.getKey()], getMLFAStatePair(p.getNonterminal2()));
                                }
                            }

                            public void visitAutomatonProduction(Nonterminal a, AutomatonProduction p) {
                            	//System.out.println("LEFT MLAF AutomatonPRODUCTION: " + a.toString() + " prod: " + p.toString());
                                // A -> reg  =>  (S)--[reg]-->(A)
                                mlfa.addAutomatonTransition(c.getState(), nt_states[a.getKey()], p.getAutomaton());
                            }

                            public void visitUnaryProduction(Nonterminal a, UnaryProduction p) {
                            	//System.out.println("LEFT MLAF UnaryPRODUCTION: " + a.toString() + " prod: " + p.toString());
                                // A -> op1(X)  =>  (S)--[op1(X)]-->(A)
                                convertComponent(comp.getComponent(p.getNonterminal()), mlfa);
                                mlfa.addUnaryTransition(c.getState(), nt_states[a.getKey()], p.getOperation(), getMLFAStatePair(p.getNonterminal()));
                            }

                            public void visitBinaryProduction(Nonterminal a, BinaryProduction p) {
                            	//System.out.println("LEFT MLAF BinaryPRODUCTION: " + a.toString() + " prod: " + p.toString());
                                // A -> op2(X,Y)  =>  (S)--[op2(X,Y)]-->(A)
                                convertComponent(comp.getComponent(p.getNonterminal1()), mlfa);
                                convertComponent(comp.getComponent(p.getNonterminal2()), mlfa);
                                mlfa.addBinaryTransition(c.getState(), nt_states[a.getKey()], p.getOperation(),
                                        getMLFAStatePair(p.getNonterminal1()),
                                        getMLFAStatePair(p.getNonterminal2()));
                            }

                            public void visitEpsilonProduction(Nonterminal a, EpsilonProduction p) {
                            	//System.out.println("LEFT MLAF EpsilonPRODUCTION: " + a.toString() + " prod: " + p.toString());
                                // A -> ""  =>  (S)---e--->(A)
                                mlfa.addEpsilonTransition(c.getState(), nt_states[a.getKey()]);
                            }
                        });
                    }
                }
            }
        }
    }
}
